"""$URL: svn://svn/repos/trunk/sitereaper/lib/verbose.py $
$Id: verbose.py 20952 2003-03-05 16:51:11Z dbinger $

The provided Verbose class can be used as a mixin to provide
a uniform method for sending extra output to stderr.

"""

import sys
from time import time, localtime, strftime

class Verbose:
    """
    Instance attributes:
      __verbose : int
        Higher values produce more output.
      __file : file
        file to write log messages to

    Use this class as a mixin.  Use set_verbose() to establish a verbose
    level, and then call note() to write output at different verbose
    levels.
    """

    def __init__ (self, verbose=1, file=None):
        self.__verbose = verbose
        self.__file = file or sys.stdout

    def set_file (self, file):
        self.__file = file

    def get_verbose (self):
        return self.__verbose

    def set_verbose (self,val):
        self.__verbose=val

    def note(self, level, *args):
        """note(level : int, args : [ any ] )

        When the verbose level is greater than or equal to level,
        write all args to stdout

        this prints a message including level and
        all args to stderr."""
        write = self.__file.write
        flush = self.__file.flush

        if self.__verbose >= level:
            now = time()
            timestamp = (strftime("%H:%M:%S", localtime(now)) +
                         ("%.3f" % (now % 1))[1:])
            write("[%s] " % timestamp)
            for a in args:
                write(str(a))
            write('\n')
            flush()



"""$URL: svn://svn/repos/trunk/sitereaper/lib/page_cache.py $
$Id: page_cache.py 20952 2003-03-05 16:51:11Z dbinger $

   Provides PageCache class.
   A PageCache caches Pages keyed by URL so that we avoid
   retrieving the same page repeatedly.
"""
import sys, os
import time, atexit, shutil, md5
from struct import unpack
from urlparse import urlparse, urlunparse
from distutils.dir_util import mkpath

from sitereaper.verbose import Verbose
from sitereaper.utils import warn
from sitereaper.cookies import CookieJar, CookieSet
from sitereaper.page import Page, Redirect

TMP_DIR = "/tmp"


class PageCache (Verbose):
    """PageCache caches pages.
    The primary method used is get_page(), which returns a Page
    from the cache, creating a new Page if none already exists
    for the given URL.

    Instance attributes:
      page_map : { ( url : string, Page ) }
        the cache of Page objects (not necessarily fetched from
        the web yet)
      cookie_jar : CookieJar
        the collection of cookies seen in our wandering across the
        web; every time we make a request, we pull out a subset
        of the cookies in the cookie jar and pass them along
      size : int
        number of pages in the cache
    """

    def __init__ (self, verbose=0):
        """PageCache(verbose : int = 0)"""
        Verbose.__init__(self, verbose)
        #self.cache_dir = os.path.join(TMP_DIR, "sitereaper.%d" % os.getpid())
        self.cache_dir = os.path.join(TMP_DIR, "sitereaper")
        mkpath(self.cache_dir)
        urlmap_filename = os.path.join(self.cache_dir, "urlmap")
        self.urlmap = open(urlmap_filename, "w")

        # Register a cleanup function.
        atexit.register(self.cleanup)

        self.page_map = {}
        self.cookie_jar = CookieJar()
        self.size = 0

    def cleanup (self):
        self.note(5, "PageCache.cleanup: nuking %s" % self.cache_dir)
        shutil.rmtree(self.cache_dir, ignore_errors=1)

    def get_size (self):
        return self.size

    def get_page (self, orig_url, retrieve=0):
        """get_page(orig_url : string) -> Page

        The fragment part of the given url is stripped, and the
        resulting url is used as a key into the page_map.  If the
        page_map has a Page with this key (or for this key with a final
        '/' removed), the Page is returned.  Otherwise, a new Page is
        inserted into the page_map for this key, and the new Page is
        returned.
        """
        self.note(6, 'get_page: original URL = ', orig_url)
        url_bits = urlparse(orig_url)[:-1]
        url = urlunparse(url_bits + ('',))
        if url != orig_url:
            self.note(6, 'get_page: de-fragged URL = ', url)

        # Check the cache for this URL.
        if self.page_map.has_key(url):
            self.note(7, 'get_page: found page in cache')
            return self.page_map[url]

        # Not in the cache -- create a new Page object.
        self.note(7, 'get_page: page not in cache, creating Page object')
        page = Page(url, self.get_verbose())
        self.size += 1
        self.page_map[url] = page

        if retrieve:
            return self.retrieve_page(page)
        else:
            return page


    def _request_page (self, page, base_filename):
        req_cookies = self.cookie_jar.get_cookies(page.get_netloc(),
                                                  page.get_path())
        if req_cookies:
            self.note(6, "will send %d cookie(s) with request: %s"
                      % (len(req_cookies), req_cookies))
            headers = {"Cookie" : req_cookies.make_request_header()}
        else:
            headers = None

        try:
            if base_filename:
                ok = page.retrieve(headers=headers,
                                   base_filename=base_filename)
            else:
                ok = page.check(headers=headers)
        except Redirect, redirect:
            # Start all over again with new URL (new cache filename,
            # new cookies, etc.)
            new_page = self.get_page(redirect.url)
            if base_filename:
                return self.retrieve_page(new_page)
            else:
                return self.check_page(new_page)

        if ok:
            cookie_hdr = page.response.getheader("Set-cookie")
            if cookie_hdr:
                self.note(7, "got a set-cookie header: %r" % cookie_hdr)
                resp_cookies = CookieSet()
                resp_cookies.parse_header(cookie_hdr,
                                          page.get_netloc(),
                                          page.get_path())
                self.note(6, "server set %d cookie(s): %s"
                          % (len(resp_cookies), resp_cookies))
                self.cookie_jar.add_cookies(resp_cookies)

        return page

    def retrieve_page (self, page):
        """retrieve_page(page : Page) -> Page

        Fetch the page content by doing an HTTP GET request.  Follows
        redirects, so the returned Page might be different from the
        requested Page (but they will be identical if there's no
        redirect).
        """
        if page.retrieved:
            # XXX what if page was a redirect?
            return page
        elif page.checked:
            warn("retrieving page %s after previously checking it" % page)

        # Compute a cache base filename for this URL by taking the MD5
        # checksum of the URL.  (The Page will add an extension to this
        # base filename once it knows the content-type.)
        md5sum = md5.new(page.url).digest()
        basename = "".join(["%08x" % i for i in unpack(">LLLL", md5sum)])
        self.urlmap.write("%s %s\n" % (basename, page.url))
        base_filename = os.path.join(self.cache_dir, basename)

        # And do the HTTP request.
        return self._request_page(page, base_filename)

    def check_page (self, page):
        """check_page(page : Page) -> Page

        Check the URL of 'page' by doing an HTTP HEAD request.  Handles
        redirects just like retrieve_page(), ie. the returned Page will
        be the passed-in object unless there is a redirect involved.
        """
        if page.checked:
            # XXX again, what if page was a redirect?
            return page
        else:
            return self._request_page(page, None)


    def has_page (self, url):
        return self.page_map.has_key(url)

    def page_checked (self, url):
        map = self.page_map
        return map.has_key(url) and map[url].checked

    def page_retrieved (self, url):
        map = self.page_map
        return map.has_key(url) and map[url].retrieved

    def get_urls (self):
        return self.page_map.keys()

    def get_urls_pages (self):
        return self.page_map.items()

    def get_pages (self):
        return self.page_map.values()





"""$URL: svn://svn/repos/trunk/sitereaper/lib/reaper.py $
$Id: reaper.py 20952 2003-03-05 16:51:11Z dbinger $

   Provides Reaper, a class for traversing web sites and
   reporting on the results.
   Provides PageParser, a class used by Reaper to
   parse html pages.
"""

import sys
import posixpath
from cPickle import dump
from types import ListType, StringType, FileType
from urlparse import urlparse, urlunparse
from time import time, localtime, strftime, timezone

from sitereaper.verbose import Verbose
from sitereaper.email_validator import EmailValidator
from sitereaper.page_cache import PageCache
from sitereaper.link_report import LinkReport


def timestamp ():
    return strftime("%Y-%m-%d %H:%M:%S", localtime(time()))

class Reaper (Verbose):
    """A Reaper manages a traversal of a web tree, parsing
    pages, checking the spelling of html files, checking
    all links, and generating a report.

    Instance attributes:
      local_urls : [string]
        a list of URLs to help identify pages that will get extra
        scrutiny.  Any page with an initial substring that matches one
        of the local_urls will get a chance to be parsed, unless it also
        has matches one of excluded_urls
      excluded_urls : [string]
        list of URLs to exclude from further scrutiny (any URL
        with a prefix in excluded_urls is excluded)
      deprecated_urls : [string]
        list of URLs to warn about if seen in the wild
      ignore_url_schemes : [string]
        list of URL schemes to ignore

      check_external : boolean = true
        enable checking external URLs with a HEAD request
      max_errors : int = None
        stop checking after this many "internal server errors" (5xx
        response code) from a local URL.  If None (default), no limit.
      _num_errors : int
        internal counter: how many server errors we've seen so far
      spellcheck_skiplist : [string]
        skip spellchecking any document whose URL starts with one of the
        strings in this list.  If the list is empty, will spellcheck the
        non-tag text of every HTML document seen from a "local" URL; if
        the list contains the empty string, no documents will be
        spellchecked.
      validation_skiplist : [string]
        skip HTML validation of any document whose URL starts with one
        of the strings in this list.  Like with spellcheck_skiplist, []
        means validate all local documents, and [''] means validate
        nothing.
      email_validator : EmailValidator
        handles validating email addresses found in mailto: links
      page_cache : PageCache
        Used for contacting and retrieving pages.
      link_report : LinkReport
        Records links from and to each page.
      spell_dictionary : string
        Name of file containing extra spelling dictionary.
        Format is one word per line.

      report_file : file


    Besides the constructor, the most useful methods here
    are traverse() and report().
    """

    def __init__ (self,
                  local_urls=None,
                  excluded_urls=None,
                  deprecated_urls=None,
                  ignore_url_schemes=None,
                  report_file=None,
                  spell_dictionary=None,
                  verbose=0):
        """Reaper(local_urls : [string] = None,
                  excluded_urls : [string] = None,
                  deprecated_urls : [string] = None,
                  report_file : file = sys.stdout,
                  spell_dictionary : string = None,
                  verbose : int = 0)
        """
        assert local_urls is None or type(local_urls) is ListType, \
               "bad local_urls: %r" % local_urls
        assert excluded_urls is None or type(excluded_urls) is ListType, \
               "bad excluded_urls: %r" % excluded_urls
        assert deprecated_urls is None or type(deprecated_urls) is ListType, \
               "bad deprecated_urls: %r" % deprecated_urls
        assert report_file is None or type(report_file) is FileType, \
               "bad report_file: %r" % report_file
        assert (spell_dictionary is None
                or type(spell_dictionary) is StringType), \
               "bad spell_dictionary: %r" % spell_dictionary

        Verbose.__init__(self, verbose)
        self.local_urls = local_urls
        self.excluded_urls = excluded_urls
        self.deprecated_urls = deprecated_urls
        self.ignore_url_schemes = ignore_url_schemes or []
        self.report_file = report_file or sys.stdout
        self.check_external = 1
        self.max_errors = None
        self._num_errors = None
        self.spellcheck_skiplist = []
        self.validation_skiplist = []
        self.email_validator = EmailValidator()
        self.page_cache = PageCache(verbose)
        self.link_report = LinkReport()
        self.spell_dictionary = spell_dictionary


    # -- Trivial modifiers ---------------------------------------------

    def disable_external_checks (self):
        self.check_external = 0

    def set_max_errors (self, max):
        self.max_errors = max

    def set_spellcheck_skiplist (self, skiplist):
        self.spellcheck_skiplist = skiplist

    def set_validation_skiplist (self, skiplist):
        self.validation_skiplist = skiplist


    # -- Trivial delegate methods --------------------------------------

    def is_local (self, page):
        return page.url_matches(self.local_urls)

    def is_excluded (self, page):
        return page.url_matches(self.excluded_urls)

    def is_deprecated (self, page):
        return page.url_matches(self.deprecated_urls)

    def get_page(self, url):
        return self.page_cache.get_page(url)

    def should_spellcheck (self, page):
        return not page.url_matches(self.spellcheck_skiplist)

    def should_validate (self, page):
        return not page.url_matches(self.validation_skiplist)


    # -- Traversal methods ---------------------------------------------

    def traverse (self, roots):
        """traverse(roots : [string], out : file)

        'roots' is a list of URLs to use as starting points.

        A report is generated and written to 'out'.
        """

        # If no local URLs provided, derive a list from the roots.
        if not self.local_urls:
            self.local_urls = []
            for url in roots:
                (scheme, netloc, path, _, _, _) = urlparse(url)
                path = posixpath.dirname(path)
                url = urlunparse((scheme, netloc, path, '', '', ''))
                self.local_urls.append(url)

        start_time = time()
        self.report("""\
sitereaper traversal started at %s

traversal parameters:
  roots:         %s
  local_urls:    %s
  excluded_urls: %s
  deprecated:    %s

dictionary: %s
""" % (timestamp(),
       ", ".join(roots),
       ", ".join(self.local_urls),
       ", ".join(self.excluded_urls),
       ", ".join(self.deprecated_urls),
       self.spell_dictionary))

        already_processed = {}          # maps Page objects to 1
        self._num_errors = 0

        self.note(5, 'roots = ', str(roots))
        for root in roots:
            self.note(1, "traversing pages from %s" % root)
            remaining = [self.page_cache.get_page(root, retrieve=1)]

            while remaining:
                page = remaining.pop(0)
                if already_processed.has_key(page):
                    continue

                self.note(3, "examining %s (%d pages in queue)"
                             % (page, len(remaining)))

                assert page.retrieved, "page %s not retrieved" % page

                if page.page_ok() and page.get_type() == 'text/html':
                    more_pages = self.traverse_page(page)
                    remaining.extend(more_pages)

                self.note(3, "generating report for %s" % page)
                self.report_page(page)
                page.cleanup()
                already_processed[page] = 1

        self.report_all_suspicious_words()

    def traverse_page (self, page):
        """traverse_page(page : Page)
           -> [Page]

        Examine the Page, follow links, check spelling, and print a
        report.

        Return a list of Pages found in links that might need more
        examination.
        """
        self.note(3, 'traversing ', page)
        page.parse()
        if self.should_spellcheck(page):
            page.spellcheck(self.spell_dictionary)
        if self.should_validate(page):
            page.validate()
        self.check_images(page)
        remaining = self.follow_links(page)
        return remaining

    def follow_links (self, page):
        """follow_links(page : Page) -> [ Page ]

        Return a list of Pages for the hrefs of anchor tags
        in the given page.
        """
        self.note(4, "following links in %s" % page)
        remaining = []
        for link_url in page.get_links():
            scheme = link_url.split(":", 1)[0]
            if scheme == "http":
                link_page = self.get_page(link_url)
                self.link_report.record_link(page.url, link_url)
                self.examine_link (page, link_page, remaining)
            elif scheme == "mailto":
                link_url = link_url.replace("&#64;", "@")
                addr = link_url[7:]
                self.note(4, "validating email address %s" % addr)
                if not self.email_validator.check(addr):
                    err = self.email_validator.get_error(addr)
                    page.add_link_warning(link_url, err)
            elif scheme not in self.ignore_url_schemes:
                page.add_link_warning(link_url, "unknown URL type")

        return remaining

    def examine_link (self, source_page, link_page, remaining):
        """examine_link(source_page : Page,
                        link_page : Page,
                        remaining : [Page])

        Examine a single link found in source_page.  Skips link
        (with appropriate logging or warning) if it is excluded
        or deprecated; calls follow_link() otherwise.
        """
        if self.is_excluded(link_page):
            self.note(5, "skipping %s (excluded)" % link_page)

        elif self.is_deprecated(link_page):
            source_page.add_link_warning(link_url, "deprecated")

        else:
            self.follow_link(source_page, link_page, remaining)

    def follow_link (self, source_page, link_page, remaining):
        """follow_link(source_page : Page,
                       link_page : Page,
                       remaining : [Page])

        Try to follow a link from 'source_page' that is not explicitly
        excluded or deprecated.  If it's local, retrieve the page
        and add it to 'remaining' (which ends up being appended to
        the master page queue); if it's external, either check it or
        skip it as appropriate.
        """
        if self.is_local(link_page):
            if not link_page.retrieved:
                # XXX it might be better just to check link_page now;
                # this would let us avoid retrieving potentially large
                # non-HTML files.  The downside is that we would do both
                # HEAD and GET on all local HTML files.
                self.note(5, "retrieving %s" % link_page)
                rpage = self.page_cache.retrieve_page(link_page)
                if rpage.is_server_error():
                    self._num_errors += 1
                    if self.max_errors and self._num_errors >= self.max_errors:
                        raise RuntimeError("too many server errors")

                if rpage is not link_page: # redirected
                    self.follow_link(source_page, rpage, remaining)
                else:
                    self.note(4, "queuing %s" % rpage)
                    remaining.append(rpage)

        elif self.check_external:
            self.note(5, "checking %s" % link_page)
            rpage = self.page_cache.check_page(link_page)

        else:
            self.note(5, "skipping %s (external)" % link_page)
            return

        # Error retrieving or checking link_page -- save a warning
        # in source_page.
        if not link_page.page_ok():
            source_page.add_link_warning(link_page.url, link_page.get_error())


    def check_images (self, page):
        """check_images(page : Page)

        """
        self.note(4, "checking images in %s" % page)
        for image_url in page.get_images():
            image_page = self.get_page(image_url)
            self.note(5, "checking image %s" % image_page)
            self.page_cache.check_page(image_page)
            if not image_page.page_ok():
                page.add_link_warning(image_url,
                                      image_page.get_error())

            self.link_report.record_link(page.url, image_url)


    # -- Per-page reporting --------------------------------------------

    def report (self, *args):
        if not args or len(args) == 1 and not args[0]:
            return
        for arg in args:
            self.report_file.write(arg)
        if args[-1][-1] != "\n":
            self.report_file.write("\n")
        self.report_file.flush()

    def report_page (self, page):
        """report_page(page : Page, root : string)

        Print a report of any problems (network error, HTTP error, bad
        links, bad words, bad HTML, ...)  found in 'page' to the report
        file.
        """

        errors = []
        if page.page_ok():
            if page.parse_error:
                errors.append("HTML parse error: %s" % page.parse_error)

            link_warnings = ["link: " + warning
                             for warning in page.format_link_warnings()]
            errors.extend(link_warnings)

            words = page.get_suspicious_words()
            if words:
                errors.extend(["bad word: %s" % word
                               for word in words])
            vresults = page.get_validation_results()
            if vresults:
                errors.extend(vresults)
        else:
            status = page.get_response_status()
            # 3xx or 4xx responses will have been reported with the
            # page that links to this one, so don't report them again.
            # 5xx responses and network errors will be reported twice --
            # that's deliberate!
            if status is None or 500 <= status <= 599:
                errors.append(page.get_error())

        if errors:
            size = page.get_size()
            time = page.get_elapsed_time()
            stats = []
            if size is not None:
                stats.append("%d bytes" % size)
            if time is not None:
                stats.append("%0.2f sec" % time)
            if stats:
                stats = " (%s)" % (", ".join(stats))
            else:
                stats = ""

            self.report(page.url + stats)
            for error in errors:
                self.report("  " + error)


    # -- After-the-fact reporting --------------------------------------

    def get_suspicious_words(self):
        """get_suspicious_words() -> { word : string  -> urls : [ string ] }

        Returns a dictionary mapping each suspicious word to a list
        of the URLs that contain the word."""
        r = {}
        for (url, page) in self.page_cache.get_urls_pages():
            pwords = page.get_suspicious_words()
            if pwords:
                for word in pwords:
                    r.setdefault(word, []).append(url)

        self.note(7, 'suspicious words=', ", ".join(r.keys()))
        return r

    def report_all_suspicious_words (self):
        """report_all_suspicious_words() -> string

        Writes a report listing all suspicious words
        found in all pages.
        """
        word_urls = self.get_suspicious_words()
        words = word_urls.keys()
        words.sort()

        if not words:
            return

        self.report("\nMisspelled or unrecognized words:")
        for word in words:
            self.report("  %s:" % word)
            for url in word_urls[word]:
                self.report('    %s' % url)

    def report_stats (self):
        s = 'Search Statistics:\n'
        s += 'Date: %s\n' % timestamp()
        s += 'Pages in Cache: %s\n\n' % str(self.page_cache.get_size())
        return s

    def report_links (self, filename):
        file = open(filename, "w")
        dump(self.link_report, file, 1)





#!/usr/bin/env python

"""$URL: svn://svn/repos/trunk/sitereaper/lib/spell.py $
$Id: spell.py 20952 2003-03-05 16:51:11Z dbinger $

A class that wraps a call to ispell and collects
the results in a dictionary that records the frequency
for each suspicious word.

"""


__revision__ = "$Id: spell.py 20952 2003-03-05 16:51:11Z dbinger $"

import sys, os
from sitereaper.utils import _exec, warn


class Speller:
    """
    A Speller wraps a call to ispell.

    Instance attributes:
      results : { string : int }
      ispell_stdout : file
      ispell_pid : int
      cmd : [string]
    """

    def __init__ (self, personal_dict=None):
        self.results = {}
        self.ispell_stdout = None
        self.ispell_pid = None
        self.cmd = ['ispell', '-l', '-W', '2']
        if personal_dict:
            self.cmd.extend(['-p', personal_dict])

    def start (self, filename):
        file = os.open(filename, os.O_RDONLY)
        (stdout_r, stdout_w) = os.pipe()
        pid = os.fork()
        if pid == 0:                    # in the child
            os.dup2(file, 0)            # connect stdin to file to spellcheck
            os.dup2(stdout_w, 1)        # connect stdout to parent
            os.close(stdout_r)
            _exec(self.cmd)
        else:                           # in the parent
            os.close(file)
            os.close(stdout_w)
            self.ispell_pid = pid
            self.ispell_stdout = os.fdopen(stdout_r, "r")

    def print_results (self,f=sys.stdout):
        results = self.get_results()
        words = results.keys()
        words.sort()
        for w in words:
            f.write("%s: %d\n" % (w, results[w]))

    def add_word (self, word):
        if not word:
            return
        if self.results.has_key(word):
            self.results[word] += 1
        else:
            self.results[word] = 1

    def clear_results (self):
        self.results = {}

    def get_results (self):
        output = self.ispell_stdout.read()
        self.ispell_stdout.close()
        (pid, status) = os.waitpid(self.ispell_pid, 0)
        if status != 0:
            warn("ispell exited with status %d" % status)

        for word in output.split('\n'):
            self.add_word(word)
        return self.results

if __name__=='__main__':
    import sys, os
    from urllib import urlretrieve
    from cStringIO import StringIO
    from sitereaper.page_parser import PageParser

    parser = PageParser()
    textfile = StringIO()

    for name in sys.argv[1:]:
        (filename, headers) = urlretrieve(name)
        file = open(filename)
        #txt_filename = filename + ".txt"
        txt_filename = "tmp.txt"
        txt_file = open(txt_filename, "w")

        parser.reset()
        parser.set_text_file(txt_file)
        parser.feed(file.read())
        parser.close()
        file.close()
        txt_file.close()

        speller = Speller("dictionaries/default")
        speller.start(txt_filename)
        print "url: %s" % name
        print "tmp file: %s" % filename
        print "txt file: %s" % txt_filename
        speller.print_results()
        #os.unlink(txt_filename)



"""$URL: svn://svn/repos/trunk/sitereaper/lib/email_validator.py $
$Id: email_validator.py 20952 2003-03-05 16:51:11Z dbinger $

Verify email addresses.  Does pretty much the right thing: look for MX
hosts, and connect to them in turn until we get through to one, and do
MAIL FROM, RCPT TO, QUIT.  Doesn't distinguish between temporary and
permanent errors, and doesn't try another MX once we have connected to
one.
"""

from popen2 import Popen3
from smtplib import SMTP
from socket import error as SocketError
from timeoutsocket import Timeout
from sitereaper.utils import warn

class SMTPError (Exception):
    def __init__ (self, code, msg):
        self.code = code
        self.msg = msg

    def __str__ (self):
        return "%d %s" % (self.code, self.msg)


class EmailValidator:

    def __init__ (self, sender=None):

        self.sender = sender or "<>"

        # DNS cache: dict mapping domain name to list of
        # (priority, mx_host) tuples, sorted by priority.
        self.mx_hosts = {}

        # Cache results of previous validations; maps whole email
        # address to boolean -- true means address OK, false means not.
        self.addr_result = {}

        # Record error message for bad addresses.
        self.addr_error = {}


    def get_mx_hosts (self, domain):
        hosts = self.mx_hosts.get(domain)
        if hosts is None:               # not in cache
            dnsmx = Popen3(["dnsmx", domain])
            dnsmx.tochild.close()
            hosts = []
            for line in dnsmx.fromchild:
                (pri, host) = line.strip().split()
                pri = int(pri)
                hosts.append((pri, host))
            dnsmx.fromchild.close()

            status = dnsmx.wait()
            if status != 0:
                warn("dnsmx exited with status %d" % status)
                return []

            hosts.sort()
            self.mx_hosts[domain] = hosts

        return [host for (pri, host) in hosts]


    def check (self, addr):
        if self.addr_result.has_key(addr):
            return self.addr_result[addr]

        (local_part, domain) = addr.split("@")
        mx_hosts = self.get_mx_hosts(domain)
        if not mx_hosts:
            self.addr_error[addr] = "no MX hosts for domain %s" % domain
            self.addr_result[addr] = 0
            return 0

        host_err = {}

        connected = 0
        for host in mx_hosts:
            try:
                smtp = SMTP(host)
            except (SocketError, Timeout), err:
                warn("error connecting to %s: %s" % (host, err[-1]))
                host_err[host] = err[-1]
                continue

            connected = 1
            try:
                (code, msg) = smtp.ehlo()
                if code != 250:
                    raise SMTPError(code, msg)
                (code, msg) = smtp.mail(self.sender)
                if code != 250:
                    raise SMTPError(code, msg)
                (code, msg) = smtp.rcpt(addr)
                if code not in (250, 251):
                    raise SMTPError(code, msg)
                smtp.quit()
            except SMTPError, err:
                self.addr_error[addr] = str(err)
                self.addr_result[addr] = 0    # failure
                break
            else:
                self.addr_result[addr] = 1    # success
                break

        if not connected:
            # tried all MX hosts, couldn't connect to any
            err = ["%s: %s" % (host, host_err[host])
                   for host in mx_hosts]
            err = "; ".join(err)
            self.addr_error[addr] = err
            self.addr_result[addr] = 0

        return self.addr_result[addr]

    def get_error (self, addr):
        return self.addr_error.get(addr)


if __name__ == "__main__":
    import os, sys
    from timeoutsocket import setDefaultSocketTimeout

    prog = os.path.basename(sys.argv[0])
    args = sys.argv[1:]
    if len(args) < 2:
        sys.exit("usage: %s sender recip [...]" % prog)

    setDefaultSocketTimeout(15)

    sender = args.pop(0)
    checker = EmailValidator(sender)
    for addr in args:
        print "** checking %s" % addr
        r = checker.check(addr)
        print "result: %r (error: %r)" % (r, checker.get_error(addr))
        print



"""$URL: svn://svn/repos/trunk/sitereaper/lib/validate.py $
$Id: validate.py 20952 2003-03-05 16:51:11Z dbinger $

Provides code to wrap nsgmls, for validation of HTML and XHTML
documents.
"""
import os, re
from sitereaper.utils import warn, _exec

# Need to recognize an XML declaration line in order to ignore it.
# W3C strongly encourages use of XML declaration lines in XHTML
# documents (see XHTML1 spec, 3.1.1).
xml_decl_re = re.compile(r'<\?xml[^>]*\?>', re.IGNORECASE)

# The document type declaration line is essential -- it's how
# we distinguish HTML 4 from XHTML.  A document without this
# line cannot be validated.
doctype_decl_re = re.compile(
    r'\<\!DOCTYPE \s+ HTML \s+ PUBLIC \s+'  # fixed stuff
    r'\"([^\"]+)\"'                         # document type
    r'(?: \s+ \"([^\"]+)\")? \s*>',         # DTD URI (I think)
    re.VERBOSE|re.IGNORECASE)

# Doctypes for HTML 4.0, 4.01, and XHTML 1.0:
#  -//W3C//DTD HTML 4.0//EN
#  -//W3C//DTD HTML 4.0 Transitional//EN
#  -//W3C//DTD HTML 4.0 Frameset//EN
#  -//W3C//DTD HTML 4.01//EN
#  -//W3C//DTD HTML 4.01 Transitional//EN
#  -//W3C//DTD HTML 4.01 Frameset//EN
#  -//W3C//DTD XHTML 1.0 Strict//EN
#  -//W3C//DTD XHTML 1.0 Transitional//EN
#  -//W3C//DTD XHTML 1.0 Frameset//EN
doctype_re = re.compile(r'-//W3C//DTD (X?HTML) ([^/]+)//EN')

nsgmls_err_re = re.compile(r'\w+:([^:]+):(\d+):(\d+):(?:([EW]):)? (.*)')

# Every Validator needs to dump stdout to /dev/null, so just open it once.
DEVNULL = os.open("/dev/null", os.O_WRONLY)

EXPAND_ETYPE = {"E": "error",
                "W": "warning"}


def spawn_pipe_nsgmls (argv):
    """
    Spawn a child process writing stdout to /dev/null.  Return a tuple
      (pid, child_stderr)
    where pid is the PID of the child and child_stderr is a
    readable file object connected to the child's stderr.
    """
    (err_r, err_w) = os.pipe()
    pid = os.fork()
    if pid == 0:                        # in the child
        os.dup2(DEVNULL, 1)             # stdout to /dev/null
        os.dup2(err_w, 2)               # connect stderr to parent
        os.close(err_r)
        _exec(argv)
    else:                               # in the parent
        os.close(err_w)
        child_stderr = os.fdopen(err_r, "r", -1)
        return (pid, child_stderr)



class Validator:

    def __init__ (self, filename):
        self.filename = filename
        self.doctype_error = None
        self.lang = None                # eg. "HTML" or "XHTML"
        self.version = None             # eg. "4.0 Transitional"
        self.results = []
        self._guess_doctype()

    def _guess_doctype (self):
        decl = ''
        num_lines = 0
        file = open(self.filename)
        while 1:
            line = file.readline()
            if xml_decl_re.search(line):
                continue

            decl += line
            num_lines += 1
            decl_match = doctype_decl_re.search(decl)
            if decl_match or num_lines > 5:
                break

        if decl_match:
            doc_type = decl_match.group(1)
            match = doctype_re.match(doc_type)
            if match:
                (self.lang, self.version) = match.group(1, 2)
            else:
                self.doctype_error = "invalid document type: %r" % doc_type
        else:
            self.doctype_error = "couldn't determine document type"

        file.close()

    def validate (self):
        if self.doctype_error:
            return

        # Validation command for HTML:
        #   nsgmls -E0 -wnon-sgml-char-ref foo.html > /dev/null
        # for XHTML:
        #   nsgmls -E0 -wxml -c /www/misc/sgml/xhtml.soc foo.html > /dev/null

        cmd = ["nsgmls", "-E0"]
        if self.lang == "HTML":
            cmd += ["-wnon-sgml-char-ref"]
        elif self.lang == "XHTML":
            cmd += ["-wxml", "-c", "/www/misc/sgml/xhtml.soc"]
        else:
            # doctype_re guards against this
            raise RuntimeError(
                "this can't happen! (self.lang == %r)" % self.lang)
        cmd.append(self.filename)

        (pid, errors) = spawn_pipe_nsgmls(cmd)
        while 1:
            line = errors.readline()
            if not line:
                break
            match = nsgmls_err_re.match(line)
            if not match:
                warn("couldn't parse nsgmls error message: %r" % line)
                continue
            (filename, line, col, etype, msg) = match.groups()
            # nsgmls has been known to issue warnings about the
            # various SGML bureaucracy files (catalogs, DTDs, etc.)
            if filename == self.filename:
                self.results.append((line, col, etype, msg))

        errors.close()
        (pid, status) = os.waitpid(pid, 0)

        # nsgmls exits with status 1 (256 in 16-bit status word)
        # when it finds errors, so ignore that status code.
        if status not in (0, 256):
            warn("nsgmls exited with status %d" % status)

    def format_results (self):
        if self.doctype_error:
            return ["error: " + self.doctype_error]
        else:
            results = []
            for (line, col, etype, msg) in self.results:
                if etype:
                    etype = EXPAND_ETYPE.get(etype) + ": "
                else:
                    etype = ""
                results.append("line %s, col %s: %s%s"
                               % (line, col, etype, msg))
        return results

if __name__ == "__main__":
    import sys
    from urllib import urlretrieve

    for name in sys.argv[1:]:
        (filename, headers) = urlretrieve(name)
        val = Validator(filename)
        val.validate()
        errors = val.format_results()
        if errors:
            print "%s:" % name
            for r in errors:
                print "  " + r



"""$URL: svn://svn/repos/trunk/sitereaper/lib/page_parser.py $
$Id: page_parser.py 20952 2003-03-05 16:51:11Z dbinger $

Simple HTML parser that finds two kinds of links ("A HREF" and "IMG
SRC"), and stores non-tag text for later spell-checking.
"""

import re
from sitereaper.verbose import Verbose
from sitereaper.utils import warn
from HTMLParser import HTMLParser

HTML_ENTITY = { "quot"    : '"',
                "amp"     : '&',
                "lt"      : '<',
                "gt"      : '>',
                "nbsp"    : ' ',
                "copy"    : '(C)',
                "reg"     : '(R)',
                "micro"   : 'u',        # as in micrometres
                "Aring"   : 'A',        # as in Angstroms
                "Eacute"  : 'E',
                "eacute"  : 'e',
                "Omega"   : 'ohm',      # ugh!
                "deg"     : 'deg',

              }

url_re = re.compile(r'(https?|ftp)://\S+')
# space allowed around the '@' because of how our /feedback page
# formats the recipient email address
email_re = re.compile(r'[\w\!\#\$\%\&\'\*\+\-\/\=\?\^\`\{\|\}\~\.]+'
                      r'\s*\@\s*'
                      r'[\w\!\#\$\%\&\'\*\+\-\/\=\?\^\`\{\|\}\~\.]+')

# A "numword" is a gibberish combination of letters and digits which we
# do not want to spellcheck, eg. PXC200, 200SL, NT3300, nght2html.
numword_re = re.compile(r'([a-zA-Z]*? \d+ [a-zA-Z]+ | '
                        r' [a-zA-Z]+? \d+ [a-zA-Z]*)', re.VERBOSE)

# Also want to skip spellchecking TLAs (three-letter acronyms) and
# XTLAs (extended three-letter acronyms, aka four-letter acronyms)
tla_re = re.compile(r'\b[A-Z]{3,4}\b')


class PageParser (HTMLParser, Verbose):
    """
    Instance attributes:
      __links : [string]
        list of URLs found in "<A HREF=...>"
      __images : [string]
        list of URLs found in "<IMG SRC=...>"
      __text_file : file
        file to which all non-tag text found in the document will
        be written.
      __suppress_text : boolean = false
        when true, non-tag text will be skipped (not written to
        __text_file) -- used for <code> and <pre> tags
    """

    SUPPRESS_TEXT_TAGS = ("pre", "code", "script", "style")

    def __init__ (self, verbose=0):
        Verbose.__init__(self, verbose)
        HTMLParser.__init__(self)

    def reset (self):
        HTMLParser.reset(self)
        self.__links = []               # list of URLs (<A HREF> links)
        self.__images = []              # list of URLs (<IMG SRC> links)
        self.__text_file = None
        self.__suppress_text = 0

    def warn (self, msg):
        (line, offset) = self.getpos()
        text = []
        if line is not None:
            text.append("line %d" % line)
        if offset is not None:
            text.append("columnd %d" % offset)
        if text:
            text = ", ".join(text) + ": "
        else:
            text = ""
        warn(text + msg)

    def set_text_file (self, file):
        self.__text_file = file

    def _write_text (self, text):
        if not self.__suppress_text:
            self.__text_file.write(text)

    def handle_starttag (self, tag, attrs):
        self.note(10, "starttag: %r" % tag)
        if tag == "a":                  # looking for "<A HREF ...>"
            for (attr, val) in attrs:
                if attr == "href":
                    self.note(7, "adding link %r" % val)
                    self.__links.append(val)

        elif tag == "img":              # looking for "<IMG SRC ...>"
            for (attr, val) in attrs:
                if attr == "src":
                    self.note(7, "adding image %r" % val)
                    self.__images.append(val)

        elif tag in ("sup", "td", "th"):
            # Turning <sup> into a space is a nasty little hack to
            # ensure that "X<sup>sm</sup>" is treated as
            # "X sm".
            self._write_text(" ")
        elif tag in ("br", "tr"):
            self._write_text("\n")

        elif tag in self.SUPPRESS_TEXT_TAGS:
            self.__suppress_text += 1

    def handle_endtag (self, tag):
        if tag in self.SUPPRESS_TEXT_TAGS:
            if self.__suppress_text == 0:
                self.error("unbalanced tag: </%s>" % tag)
            self.__suppress_text -= 1
        elif tag in ("h1", "h2", "h3", "h4", "h5", "h6"):
            self._write_text("\n")

    def handle_data (self, data):
        assert self.__text_file is not None, "text_file not set"

        if self.get_verbose() >= 10:
            if len(data) > 50:
                self.note(10, "data: '%s ... %s'" % (data[0:40], data[-5:]))
            else:
                self.note(10, "data: %r" % data)

        data = url_re.sub('', data)
        data = email_re.sub('', data)
        data = numword_re.sub('', data)
        data = tla_re.sub('', data)

        self._write_text(data)

    def handle_entityref (self, name):
        text = HTML_ENTITY.get(name)
        if text is None:
            self.warn("HTML entity &%s; not recognized" % name)
        else:
            self._write_text(text)

    def handle_charref (self, ref):
        try:
            ref = int(ref)
        except ValueError:
            self.warn("bogus HTML character reference &#%s;" % ref)
            return
        else:
            if 32 <= ref <= 127:
                self._write_text(chr(ref))

    def get_links (self):
        return self.__links

    def get_images (self):
        return self.__images


if __name__ == "__main__":
    import sys
    infile = sys.argv[1]

    file = open(infile)
    parser = PageParser(verbose=0)
    parser.set_text_file(sys.stdout)
    print "-- parsing %s ---------" % infile
    parser.feed(file.read())
    file.close()
    parser.close()
    print "--------------------------------"

    print "links:"
    for url in parser.get_links():
        print "  " + url
    print "images:"
    for url in parser.get_images():
        print "  " + url



"""$URL: svn://svn/repos/trunk/sitereaper/lib/page.py $
$Id: page.py 20952 2003-03-05 16:51:11Z dbinger $

Provides the Page class, used to store everything we know about
a single web resource (URL, HTTP response, content, etc.).
"""

import sys, os
import time
import socket
from urlparse import urlparse, urlunparse, urljoin
from httplib import HTTPConnection
from HTMLParser import HTMLParseError
from timeoutsocket import Timeout
import sitereaper
from sitereaper.utils import std_repr, warn
from sitereaper.verbose import Verbose
from sitereaper.spell import Speller
from sitereaper.validate import Validator
from sitereaper.page_parser import PageParser


FILE_EXTENSION = {
    'text/plain' : '.txt',
    'text/html'  : '.html',
    'image/jpeg' : '.jpeg',
    'image/gif' : '.gif',
    'image/png' : '.png',
    }



# Used by Page to inform PageCache of a redirect.
class Redirect (Exception):
    def __init__ (self, status, reason, url):
        self.status = status
        self.reason = reason
        self.url = url

    def __str__ (self):
        return "%d %s: %s" % (self.status, self.reason, self.url)


class Page(Verbose):
    """
    A Page represents a web resource (ie. the object at the other
    end of an HTTP URL), and everything that went wrong while
    retrieving or inspecting it.  This can include:
      * network problems (hostname didn't resolve, connection
        refused or timed out)
      * HTTP problems (3xx, 4xx, 5xx response) (note that a 3xx
        response isn't really a problem, but if some other Page
        links to a URL that just does a redirect, we want to
        know about it)
      * bad links (if it's an HTML document, it links to some
        resource that had either network or HTTP problems)
      * bad text (eg. spell-checking or HTML validation found
        problems with the content of this Page -- again only for
        HTML docs)


    Instance attributes:
      url : string
        the URL given in the constructor, with the fragment stripped
      parsed_url : ( scheme : string, \
                     netloc : string, \
                     path : string, \
                     params : sting, \
                     query : string, \
                     fragment : string )
        <scheme>://<netloc>/<path>;<params>?<query>#<fragment>

      network_error : string
        string containing any error message that resulted from
        attempting to connect to the HTTP server; None means
        no error
      response : httplib.HTTPResponse
        the response received from the server when this page was
        requested

      filename : string
        name of the file to which the page's content was written;
        None if we didn't save the content of the page (eg. on error
        or redirect)
      text_filename : string
        name of file with page's non-tag text -- only used for
        HTML documents
      parser : PageParser
        A PageParser that has been used to parse the contents of this page.
      parse_error : HTMLParser.HTMLParseError
        if there were errors parsing the page, the exception object
        that signaled the error and records the information about it
      size : int
        The number of bytes in the content for this web page.
      elapsed_time : float
        number of seconds it took to retrieve this page

      checked : boolean
        true if we have done either a HEAD or a GET for this URL
        (regardless of success)
      retrieved : boolean
        true if we have done a GET for this URL
        (regardless of success)

      spellcheck_results : { string : int }
        dictionary of words reported by ispell from the non-tag content
        of this web page; the dict records how many times each word
        occurs in the page
      validation_results : [ string ]
        list of error messages from running an HTML validator
        on the page's HTML content

      link_warnings : [(string, string)]
        a list of (url, message) tuples, one for each link in the
        page that had some sort of problem (eg. a network error,
        HTTP error, permanent HTTP redirect, unchecked link, ...)
    """

    def __init__ (self, url, verbose=0):
        """Page(url : string, verbose : int = 0)"""
        Verbose.__init__(self, verbose)
        self.set_url(url)               # sets url, parsed_url

        self.network_error = None
        self.response = None

        self.base_filename = None
        self.filename = None
        self.text_filename = None
        self.parser = None
        self.parse_error = None
        self.size = None
        self.elapsed_time = None

        self.checked = 0
        self.retrieved = 0

        self.spellcheck_results = None
        self.validation_results = None
        self.link_warnings = []

    def __str__ (self):
        return self.url

    __repr__ = std_repr

    def set_url (self, url):
        self.url = url
        self.parsed_url = urlparse(url)
        if self.parsed_url[0] != "http":
            raise ValueError, "Page class is only for HTTP URLs"

    def url_matches (self, url_prefixes):
        """url_matches(urls : [string]) -> boolean

        Return true if the URL of this page matches any of
        url_prefixes.
        """
        if url_prefixes is None:        # treat None like an empty list
            return 0
        url = self.url
        for prefix in url_prefixes:
            if url.startswith(prefix):
                return 1
        return 0

    def get_suspicious_words (self):
        if self.spellcheck_results:
            words = self.spellcheck_results.keys()
            words.sort()
            return words
        else:
            return None

    def get_validation_results (self):
        return self.validation_results

    def get_url (self): return self.url

    def get_scheme (self):   return self.parsed_url[0]
    def get_netloc (self):   return self.parsed_url[1]
    def get_path (self):     return self.parsed_url[2]
    def get_params (self):   return self.parsed_url[3]
    def get_query (self):    return self.parsed_url[4]
    def get_fragment (self): return self.parsed_url[5]

    def _make_request (self, headers, method="GET"):
        netloc = self.parsed_url[1]
        selector = urlunparse((None, None) + self.parsed_url[2:])
        try:
            conn = HTTPConnection(netloc)
            conn.putrequest(method, selector)
            conn.putheader("User-agent",
                           "sitereaper/%s" % sitereaper.__version__)
            if headers:
                for (key, value) in headers.items():
                    conn.putheader(key, value)
            conn.endheaders()
            response = conn.getresponse()
        except socket.error, err:
            self.network_error = "%s: %s" % (netloc, err[1])
            return (None, None)
        except Timeout, err:
            self.network_error = str(err)
            return (None, None)
        else:
            return (conn, response)

    def _interpret_response (self, response):
        read_content = 0
        if 100 <= response.status <= 199:
            # 1xx responses shouldn't be sent to HTTP 1.0 clients,
            # but clients are allowed to ignore them.
            ok = 1
        if 200 <= response.status <= 299:     # successful completion
            # should only get 200 responses, so warn if anything else
            if response.status != 200:
                warn("%s: unexpected success response: %s %s"
                     % (self.url, response.status, response.reason))
            read_content = 1
            ok = 1

        if 300 <= response.status <= 399:     # redirect of some sort
            if response.status not in (301, 302, 307):
                warn("%s: unexpected redirect response: %s %s"
                     % (self.url, response.status, response.reason))
            url = response.getheader("location")
            # Raise an exception because PageCache needs to cook
            # up a new cache filename for the new URL.
            self.note(5, "redirecting to %s" % url)
            raise Redirect(response.status, response.reason, url)

        elif 400 <= response.status <= 499:   # request error
            ok = 0
        elif 500 <= response.status <= 599:   # server error
            warn("%s: server error: %s %s"
                 % (self.url, response.status, response.reason))
            ok = 0

        return (ok, read_content)

    def _read_content (self, response, base_filename):

        headers = response.msg
        ext = FILE_EXTENSION.get(headers.gettype(), '')

        self.filename = filename = base_filename + ext
        self.note(8, "response received, writing content to %s" % filename)
        file = open(filename, "w")
        size = 0
        while 1:
            data = response.read(8*1024)
            if not data:
                break
            size += len(data)
            file.write(data)
        file.close()
        self.size = size

        hdr_filename = base_filename + ".hdr"
        self.note(8, "writing response headers to %s" % hdr_filename)
        hfile = open(hdr_filename, "w")
        print >>hfile, headers
        hfile.close()

    def retrieve (self, headers=None, base_filename=None):
        """retrieve(headers : { string : string } = None,
                    base_filename : string = None)

        Retrieve the content of this page and store it in a cache file.
        This should probably only be called from PageCache, to ensure
        that cookies are handled properly.

        Returns true if we successfully connected to the server, got a
        2xx response, and read the whole body.  Raises Redirect if we
        got a 3xx response.  Returns false otherwise.
        """
        self.note(7, "fetching %s" % self.url)
        start_time = time.time()
        (connection, response) = self._make_request(headers)
        self.response = response
        if connection and response:     # no network errors
            self.note(7, "request started OK, response = %s %s"
                         % (response.status, response.reason))

            try:
                (ok, read_content) = self._interpret_response(response)

                if read_content:
                    # Should only get here for a 2xx response.  Copy the
                    # response content to the cache file.
                    self.base_filename = base_filename
                    try:
                        self._read_content(response, base_filename)
                    except Timeout, err:
                        self.network_error = "%s: %s" % (self.url, err)
                        self.note(7, "network error reading content")
                        ok = 0
                    else:
                        clen = self.response.getheader("Content-length")
                        if clen and int(clen) != self.size:
                            warn("%s: content-length said %d bytes, "
                                 "but got %d bytes"
                                 % (self.url, clen, self.size))
            finally:
                self.checked = self.retrieved = 1
                connection.close()
        else:
            self.checked = self.retrieved = 1
            self.note(7, "network problem: %s" % self.network_error)
            ok = 0

        stop_time = time.time()
        self.elapsed_time = stop_time - start_time

        return ok

    def check (self, headers=None):
        self.note(7, "checking %s" % self.url)
        start_time = time.time()
        (connection, response) = self._make_request(headers, method="HEAD")
        self.response = response
        if connection and response:     # no network errors
            self.note(7, "made request, response = %s %s"
                         % (response.status, response.reason))
            try:
                (ok, _) = self._interpret_response(response)
            finally:
                self.checked = 1
                connection.close()
        else:
            self.checked = 1
            ok = 0

        self.elapsed_time = time.time() - start_time
        return ok

    def get_type (self):
        """get_type() -> string

        Return the MIME type reported when we requested this page.
        """
        if self.response:
            return self.response.msg.gettype()

    def get_size (self):
        """get_size() -> int

        Report the best information available about the number of bytes
        in the content of this web page.
        """
        return self.size

    def get_elapsed_time (self):
        return self.elapsed_time

    def get_response_status (self):
        return self.response and self.response.status

    def get_response_reason (self):
        return self.response and self.response.reason

    def get_error (self):
        if self.network_error:
            return self.network_error
        elif self.response:
            return "%03d %s" % (self.response.status, self.response.reason)
        else:
            return None

    def page_ok (self):
        return (self.network_error is None and
                self.response is not None and
                (200 <= self.response.status <= 299 or
                 self.response.status in (302, 307)))

    def is_request_error (self):
        return self.response and 400 <= self.response.status <= 499

    def is_server_error (self):
        return self.response and 500 <= self.response.status <= 599

    # XXX not thread-safe -- but who cares?
    def is_parsed (self):
        return self.parser is not None

    def parse (self):
        """Parse the content of the page: extract links and non-tag
        text for later spell-checking.
        """
        self.parser = PageParser(verbose=self.get_verbose())
        file = open(self.filename)
        self.text_filename = self.base_filename + ".txt"
        text_file = open(self.text_filename, "w")
        self.parser.set_text_file(text_file)
        try:
            self.parser.feed(file.read())
            self.parser.close()
        except HTMLParseError, err:
            self.parse_error = err
        text_file.close()
        file.close()

    def get_links (self):
        return [urljoin(self.url, url) for url in self.parser.get_links()]

    def get_images (self):
        return [urljoin(self.url, url) for url in self.parser.get_images()]

    def add_link_warning (self, url, message):
        assert message, \
               ("page %s: adding empty message %r for link %r" %
                (self, message, url))
        self.link_warnings.append((url, message))

    def format_link_warnings (self):
        return ["%s: %s" % (warning) for warning in self.link_warnings]


    def spellcheck (self, dictionary):
        speller = Speller(dictionary)
        speller.start(self.text_filename)
        self.spellcheck_results = speller.get_results()

    def validate (self):
        val = Validator(self.filename)
        val.validate()
        self.validation_results = val.format_results()

    def cleanup (self):
        try:
            if self.filename is not None and os.path.exists(self.filename):
                os.remove(self.filename)
            if self.text_filename is not None:
                os.remove(self.text_filename)
        except OSError, err:
            warn("%s: %s" % (err.filename, err.strerror))



"""$URL: svn://svn/repos/trunk/sitereaper/lib/link_report.py $
$Id: link_report.py 20952 2003-03-05 16:51:11Z dbinger $

Provides the LinkReport class, which exists mainly to make
accessing the link report pickles easier.
"""

class LinkReport:
    """

    Instance attributes:
      links_to : { string : [string] }
        maps the a URL of a page to the list of URLs that the page links to
      linked_from : { string : [string] }
        maps the URL of a page to the list of URLs of pages that link
        to that page

    """

    def __init__ (self):
        self.links_to = {}
        self.linked_from = {}

    def record_link (self, referring_url, target_url):
        self.links_to.setdefault(referring_url, [])
        self.linked_from.setdefault(target_url, [])

        self.links_to[referring_url].append(target_url)
        self.linked_from[target_url].append(referring_url)

    def report_targets (self, url, indent=""):
        for url in self.links_to.get(url, []):
            print indent + url

    def report_referrers (self, url, indent=""):
        for url in self.linked_from.get(url, []):
            print indent + url

    def report (self, url):
        print url
        print " links to:"
        self.report_targets(url, indent="  ")
        print " linked from:"
        self.report_referrers(url, indent="  ")



"""$URL: svn://svn/repos/trunk/sitereaper/lib/utils.py $
$Id: utils.py 20952 2003-03-05 16:51:11Z dbinger $

Miscellaneous utility functions needed by other sitereaper modules.
"""

import sys, os

# Pseudo-method that can be used in any class
def std_repr (self):
    if hasattr(self, '__str__'):
        return "<%s at %x: %s>" % (self.__class__.__name__, id(self), self)
    else:
        return "<%s at %x>" % (self.__class__.__name__, id(self))

def warn (msg):
    sys.stderr.write("warning: %s\n" % msg)


def _exec (argv):
    try:
        os.execvp(argv[0], argv)
    except OSError, err:
        os.write(2, "couldn't exec %s: %s\n" % (argv[0], err.strerror))
    except:
        os.write(2, "couldn't exec %s\n" % argv[0])
    os._exit(1)


def spawn_pipe_r (argv, bufsize=-1):
    """spawn_pipe_r(argv : [string], bufsize : int = -1)
       -> (pid : int, child_stdout : file)

    Spawn a child process; return the PID of the child and a readable
    file object connected to the child's stdout.  (The child's stdin and
    stderr are the same as the parent's.)
    """
    (fd_r, fd_w) = os.pipe()
    pid = os.fork()
    if pid == 0:                        # in the child
        os.dup2(fd_w, 1)                # connect stdout to the parent
        os.close(fd_r)                  # not needed here
        _exec(argv)
    else:                               # in the parent
        os.close(fd_w)
        child_stdout = os.fdopen(fd_r, "r", bufsize)
        return (pid, child_stdout)



#!/usr/bin/env python

"""$URL: svn://svn/repos/trunk/sitereaper/setup.py $
$Id: setup.py 21296 2003-04-08 18:43:43Z akuchlin $

Setup script for sitereaper
"""

import os
from distutils import core

kw = {'name': "Sitereaper",
      'version': "0.1",
      'description': "A link checker with optional HTML validation and spell-checking",
      'author': "MEMS Exchange",
      'author_email': "quixote@mems-exchange.org",
      'url': "http://www.mems-exchange.org/software/sitereaper.html",

      'package_dir': {'sitereaper':'lib'},
      'packages': ['sitereaper'],
      'scripts': ['bin/sitereaper'],
     }

# If we're running Python 2.3, add extra information
if hasattr(core, 'setup_keywords'):
    if 'classifiers' in core.setup_keywords:
        kw['classifiers'] = ['Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Operating System :: Unix',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: MacOS :: MacOS X',
          'Topic :: Internet :: WWW/HTTP :: Site Management :: Link Checking',
          ]
    if 'download_url' in core.setup_keywords:
        kw['download_url'] = ('http://www.mems-exchange.org/software/files'
                              '/sitereaper/%s-%s.tar.gz' %
                              (kw['name'], kw['version']) )

core.setup(**kw)



"""$URL: svn://svn/repos/trunk/sitereaper/lib/cookies.py $
$Id: cookies.py 20955 2003-03-05 17:05:52Z dbinger $

An attempt to correct for the brain-damaged terminology ("morsel"?!?)
and excessively complex code in Cookie.py.
"""

import Cookie
from time import time, gmtime, strftime, strptime, mktime, timezone
from sitereaper.utils import std_repr
from sitereaper.verbose import Verbose


_cookie_re = Cookie._CookiePattern

_cookie_attrs = ("comment",
                 "domain",
                 "max-age",
                 "path",
                 "secure",
                 "version",
                 "expires",           # for old (pre-RFC 2109) cookies
                )

EXPIRES_FMT = "%a, %d-%b-%Y %H:%M:%S GMT"

class Cookie:
    """
    A single cookie, i.e. a name/value pair with several associated
    attributes (domain, path, secure, etc.).  For parsing the cookie-
    related HTTP headers ("Cookie" and "Set-Cookie"), use the
    CookieSet class; to generate those cookies, you can use either this
    class or CookieSet.

    Instance attributes:
      name : string
        the name of the cookie
      value : string
        the value of the cookie, untouched (eg. if the original value
        was quoted, this includes the quotes)
      comment : string
      domain : string
      path : string
      max_age : string
        maximum age of the cookie, in seconds from the time of
        receipt by the client
      secure : boolean
      version : string
      expires : int
        expiration time of the cookie, in seconds since the epoch
        (1970-01-01 00:00 UTC).

      creation_time : int
        the time this Cookie object was created, which is presumably
        equivalent to the time the cookie was received by the
        application.  Only relevant for client-side applications.

    You can set or get the cookie attributes specified by RFC 2109
    either as instance attributes or dictionary items:

      cookie["Domain"] = ".example.com"
      print cookie.domain

      cookie["Max-age"] = 600
      print cookie.max_age

    Pseudo-dictionary keys are case insensitive.

    The "Expires" attribute is subject to special handling if set like a
    dictionary item.  Eg. this assignment:

      cookie["expires"] = "Sun, 17-Jan-2038 19:14:07 GMT"

    results in the integer value 2147386447 (seconds since epoch for
    that date) being assigned to the 'expires' attribute.  A similar
    translation occurs on dictionary access to the "expires" attribute.
    """

    def __init__ (self, name, value):
        self.name = name
        self.value = value
        self.comment = None
        self.domain = None
        self.path = None
        self.max_age = None
        self.secure = None
        self.version = None
        self.expires = None

        self.creation_time = time()

    def __str__ (self):
        return "%s = %s" % (self.name or "*no name*",
                            self.value or "*no value*")

    __repr__ = std_repr

    def __eq__ (self, other):
        return vars(self) == vars(other)

    def __ne__ (self, other):
        return vars(self) != vars(other)

    # This is just for the convenience of CookieSet.parse_header().
    def __setitem__ (self, key, val):
        key = key.lower()
        if key not in _cookie_attrs:
            raise IndexError, key
        key = key.replace("-", "_")
        if key == "expires":
            try:
                val = mktime(strptime(val, EXPIRES_FMT)) - timezone
            except ValueError:
                val = None
        elif key == "max_age":
            try:
                val = int(val)
            except ValueError:
                val = None
        setattr(self, key, val)

    def __getitem__ (self, key):
        key = key.lower()
        if key not in _cookie_attrs:
            raise IndexError, key
        key = key.replace("-", "_")
        if key == "expires" and self.expires is not None:
            return strftime(EXPIRES_FMT, gmtime(self.expires))
        else:
            return getattr(self, key)

    def get_age (self):
        """Return the age of the cookie in seconds."""
        return time() - self.creation_time

    def is_expired (self):
        """Return true if the cookie has expired."""
        # Prefer the RFC 2109 'max-age' attribute.
        if self.max_age is not None:
            return self.get_age() > self.max_age

        # Fallback to pre-RFC 2109 (Netscape) "expires"
        if self.expires is not None:
            return time() > self.expires

    def path_matches (self, path):
        if (self.path is None or self.path == "/" or self.path == path):
            return 1
        else:
            return (path.startswith(self.path) and
                    path[len(self.path)] == "/")

    def make_response_header (self):
        r = ["%s=%s" % (self.name, self.value)]
        for attr in _cookie_attrs:
            val = self[attr]
            if val:
                r.append("%s=%s" % (attr.capitalize(), val))

        return ";".join(r)

    def make_request_header (self):
        assert self.name and self.value
        if self.version is None:        # pre-RFC 2109
            return "%s=%s" % (self.name, self.value)
        else:
            return "$Version=%s; %s=%s" % (self.version, self.name, self.value)


class CookieSet:

    def __init__ (self, cookies=None):
        if cookies:
            self.cookies = cookies
        else:
            self.cookies = []

    def __str__ (self):
        if self.cookies:
            return ", ".join([cookie.name for cookie in self.cookies])
        else:
            return "none"

    __repr__ = std_repr

    def __len__ (self):
        return len(self.cookies)

    def __getitem__ (self, i):
        return self.cookies[i]

    def __nonzero__ (self):
        return len(self.cookies) > 0

    def parse_header (self, str, server, request_uri=None):
        """
        Parse an entire RFC 2109 "Set-cookie" header, resulting in
        one or more Cookie objects (stored in self.cookies).
        """
        i = 0                           # current char of str
        cookie = None                   # current cookie -- doesn't exist
                                        # until we see the cookie's name
        path = None
        if request_uri:
            lastslash = request_uri.rindex('/')
            if lastslash != -1:
                path = request_uri[0:lastslash]

        # XXX doesn't do anything about unquoting quoted cookie values.
        # NB. neither does the standard lib's Cookie.py, and it might
        # not matter if all we're doing is storing the value briefly
        # before sending it back to the server.

        while i < len(str):
            match = _cookie_re.search(str, i)
            if not match:
                break

            key = match.group("key")
            val = match.group("val")
            i = match.end(0)

            if key[0] == "$":
                pass                    # XXX what to do about these?
            elif key.lower() in _cookie_attrs:
                if cookie:
                    cookie[key.lower()] = val
                else:
                    raise ValueError(
                        "bad cookie header: attr %s before name: %r"
                        % (key, str))
            else:
                cookie = Cookie(key, val)
                cookie.domain = server  # default -- can be overridden
                cookie.path = path
                self.cookies.append(cookie)

    def make_response_header (self):
        hdrs = [cookie.make_response_header() for cookie in self.cookies]
        return ", ".join(hdrs)

    def make_request_header (self):
        hdrs = [cookie.make_request_header() for cookie in self.cookies]
        return "; ".join(hdrs)


class CookieJar (Verbose):
    """
    Client-side collection of HTTP cookies to pass to various servers
    under various conditions.
    """

    def __init__ (self):
        # map domain (server) names to a list of cookies: eg.
        #  { "cthulhu.gerg.ca" :
        #    [Cookie("userid", "gward", ...),
        #     Cookie("session", "127.0.0.1:20020130184242:....")],
        #    "google.com" :
        #    [Cookie("PREF", "...")],
        #  }
        self.cookies = {}

    __repr__ = std_repr

    # XXX misnamed, because it can remove a cookie too!
    def add_cookie (self, cookie):
        """Feed the cookie jar by adding a new cookie to it."""
        cookies = self.cookies.get(cookie.domain)
        if cookies:
            for i in range(len(cookies)-1, -1, -1):
                if (cookies[i].name == cookie.name and
                    cookies[i].path == cookie.path):
                    if cookie.max_age == 0:
                        del cookies[i]
                    else:
                        cookies[i] = cookie
        else:
            if cookie.max_age != 0:
                self.cookies[cookie.domain] = [cookie]

    def add_cookies (self, cookie_set):
        for cookie in cookie_set:
            self.add_cookie(cookie)

    def get_cookies (self, server, path):
        """get_cookies(server : string, path : string) -> CookieSet

        Determine which cookies should be sent to 'path' on 'server'
        and return a CookieSet containing them.
        """

        # First pass: get cookies that match on server/domain.
        cookies = self.cookies.get(server)
        if cookies:
            cookies = cookies[:]        # make a copy to extend
        else:
            cookies = []

        # Umm, this might be too general... "x.y.foo.com" is not
        # supposed to match ".foo.com".
        name_bits = server.split(".")[1:]
        while len(name_bits) > 1:
            n = "." + ".".join(name_bits)
            c = self.cookies.get(n)
            if c:
                cookies.extend(c)
            del name_bits[0]

        # Now go through that list, throwing away cookies that
        # don't match by path or time.
        now = time()
        for i in range(len(cookies)-1, -1, -1):
            cookie = cookies[i]
            if cookie.is_expired():
                del cookies[i]
                self.cookies[cookie.domain].remove(cookie)
            elif not cookie.path_matches(path):
                del cookies[i]

        return CookieSet(cookies)


if __name__ == "__main__":
    server = "cthulhu.gerg.ca"

    cset = CookieSet()
    cset.parse_header('userid="gward";Domain=.gerg.ca', server)
    print cset[0]
    print cset[0].make_response_header()
    print cset[0].make_request_header()
    print

    jar = CookieJar()
    jar.add_cookie(cset[0])
    cset2 = jar.get_cookies("cthulhu.gerg.ca", "/")
    print cset2[0]
    assert len(cset2) == 1 and cset2[0] == cset[0]

    cset3 = jar.get_cookies("blah.gerg.ca", "/foo")
    assert len(cset3) == 1 and cset3[0] == cset[0]

    cset4 = jar.get_cookies("foogerg.ca", "/")
    assert len(cset4) == 0

    cset = CookieSet()
    cset.parse_header('userid="gward";doMaiN=.gerg.ca; '
                      'session=ab93cnd;Path=/foo; '
                      'foo=bar; Domain=.gerg.ca; Path=/foo',
                      server)
    assert len(cset) == 3, `cset`
    print cset[0]
    print cset[0].make_response_header()
    print cset[0].make_request_header()
    print
    print cset[1]
    print cset[1].make_response_header()
    print cset[1].make_request_header()
    print
    print cset[2]
    print cset[2].make_response_header()
    print cset[2].make_request_header()
    print
    print cset.make_response_header()
    print cset.make_request_header()

    cjar = CookieJar()
    cjar.add_cookies(cset)
    print `cjar.get_cookies("cthulhu.gerg.ca", "/")` # _userid only
    print `cjar.get_cookies("cthulhu.gerg.ca", "/foo")` # both
    print `cjar.get_cookies("cthulhu.gerg.ca", "/foobar")` # userid only
    print `cjar.get_cookies("www.gerg.ca", "/")` # userid only
    print `cjar.get_cookies("www.gerg.ca", "/foo/baz")` # session only
    print `cjar.get_cookies("www.gerg.ca", "/foob/ar")`




"""$URL: svn+ssh://svn/repos/trunk/sitereaper/lib/__init__.py $
$Id: __init__.py 21763 2003-05-30 20:27:55Z dbinger $

A web crawler for spellchecking and validating HTML pages.
Catches dead links at the same time, naturally.
"""

__version__ = "0.1"
